"""
Module for the test-tool-run-process-plugin.
"""
from enum import Enum
from errno import ESRCH
from io import TextIOWrapper
from json import loads
from logging import DEBUG, debug, error, info, root
from pathlib import Path
from subprocess import PIPE, Popen
from threading import Event, Thread, Timer
from typing import Any, Dict, Hashable, List, Optional, TypedDict


class RunProcessSaveType(Enum):
    """
    This class represents the type of the save.
    """

    JSON = 1
    STRING = 2


class BashProgramType(Enum):
    """
    This class represents the type of the bash program.
    """

    BASH = "/bin/bash"
    SH = "/bin/sh"
    ZSH = "/bin/zsh"
    POWERSHELL = "powershell.exe"
    CMD = "cmd.exe"


class RunProcessSave(TypedDict):
    """
    This class represents a run process save.
    """

    name: str
    type: RunProcessSaveType
    path: Optional[List[str]]


class RunProcessCall(TypedDict):
    """
    This class represents a run process call.
    """

    cmd: str
    timeout: int
    save: Optional[RunProcessSave]
    shell: bool
    text: bool
    program: BashProgramType
    return_code: Optional[int]


append_to_run_process_call: Dict[Hashable, str] = {
    BashProgramType.BASH: "\nexit\n",
    BashProgramType.SH: "\nexit\n",
    BashProgramType.ZSH: "\nexit\n",
    BashProgramType.POWERSHELL: "\nExit\n",
    BashProgramType.CMD: "\nExit\n",
}

default_run_process_call: RunProcessCall = {
    "cmd": None,  # type: ignore
    "timeout": 0,
    "save": None,
    "shell": True,
    "text": True,
    "program": "/bin/bash",  # type: ignore
    "return_code": 0,
}


def log_output(
    pipe,
    prefix: str,
    is_error: bool,
    output_list: List,
    stop_event: Event,
):
    """
    Log the output of a pipe.

    Parameters
    ----------
    pipe : IO
        The pipe.
    prefix : str
        The prefix.
    is_error : bool
        If the pipe is an error pipe.
    output_list : List
        The output list.
    stop_event : Event
        The stop event.
    """
    if is_error:
        log = error
    else:
        log = info

    for line in iter(pipe.readline, b""):
        if line.strip():
            log(f"{prefix}: {line.strip()}")
            output_list.append(line)
        # Stop the thread if the stop event is set
        if stop_event and stop_event.is_set():
            break


def timeout_cmd(p: Popen):
    """
    Timeout a cmd.

    Parameters
    ----------
    p : Popen
        The cmd.
    """
    if p.poll() is None:
        try:
            p.terminate()
            if DEBUG >= root.level:
                error("Timeout for cmd reached.")
            exit(1)
        except OSError as e:
            if e.errno != ESRCH:
                raise


def run_cmd(
    command: str,
    timeout: float,
    shell: bool,
    text: bool,
    program: BashProgramType,
    expected_return_code: Optional[int],
) -> str:
    """
    Run a cmd.

    Parameters
    ----------
    command : str
        The command.
    timeout : float
        The timeout.
    shell : bool
        If the command should be run in a shell.
    text : bool
        If the command should be run in text mode.
    program : BashProgramType
        The program.
    expected_return_code : Optional[int]
        The expected return code.

    Returns
    -------
    str
        The output of the cmd.
    """
    process = Popen(
        program.value,
        shell=shell,
        stdout=PIPE,
        stderr=PIPE,
        stdin=PIPE,
        text=text,
    )

    if timeout:
        # Start a timer that will kill the cmd if it runs too long
        t = Timer(interval=timeout, function=timeout_cmd, args=[process])
        t.start()

    # Create an event to signal threads to stop
    stop_event = Event()
    # Start threads to read output and error streams
    output_list: List[str] = []
    output_thread = Thread(
        target=log_output,
        args=(process.stdout, "Output", False, output_list, stop_event),
    )
    error_thread = Thread(
        target=log_output,
        args=(process.stderr, "Error", True, output_list, stop_event),
    )

    output_thread.start()
    error_thread.start()

    # Stream the output of the cmd but also store it in a list
    assert isinstance(process.stdin, TextIOWrapper)
    process.stdin.write(command + append_to_run_process_call[program])
    process.stdin.flush()

    return_code = process.wait()

    # Signal the threads to stop
    stop_event.set()
    # Wait for the threads to finish
    output_thread.join()
    error_thread.join()

    if timeout:
        t.cancel()

    if (
        expected_return_code is not None
        and return_code != expected_return_code
    ):
        error(
            f"Cmd terminated with code {return_code} insted of"
            + " {expected_return_code}."
        )
        assert False

    return "".join(output_list).strip()


def make_run_process_call(call: RunProcessCall, data: Dict[str, Any]) -> None:
    """
    Make a run process call.

    Parameters
    ----------
    call : RunProcessCall
        The call.
    data : Dict[str, Any]
        The data.
    """
    info(f'Run the cmd {call["cmd"]}.')

    result = run_cmd(
        call["cmd"],
        call["timeout"],
        call["shell"],
        call["text"],
        call["program"],
        call["return_code"],
    )

    if call["save"] is not None:
        if call["save"]["type"] is None:
            call["save"]["type"] = RunProcessSaveType.STRING
        else:
            try:
                call["save"]["type"] = RunProcessSaveType[
                    call["save"]["type"]
                ]  # type: ignore
            except KeyError as e:
                raise ValueError(
                    f"Save type {str(e)} is not supported."
                ) from e

        if call["save"]["type"] == RunProcessSaveType.JSON:
            val = loads(result.replace("'", '"'))
            try:
                call["save"]["path"]
            except KeyError:
                call["save"]["path"] = []

            for p in call["save"]["path"]:  # type: ignore
                val = val[p]

            debug(f'Save {val} as {call["save"]["name"]}')
            data[call["save"]["name"]] = val
        elif call["save"]["type"] == RunProcessSaveType.STRING:
            debug(f'Save {result} as {call["save"]["name"]}')
            data[call["save"]["name"]] = result


def augment_run_process_call(
    call: RunProcessCall,
    data: Dict[str, Any],  # pylint: disable=unused-argument
    path: Path,  # pylint: disable=unused-argument
) -> None:
    """
    Augment an run process call.

    Parameters
    ----------
    call : RunProcessCall
        The call.
    data : Dict
        The data.
    path : Path
        The path.
    """
    # timeout as int
    call["timeout"] = int(call["timeout"])

    # return_code as int
    if call["return_code"] is not None:
        call["return_code"] = int(call["return_code"])

    # Default save type is string
    if call["save"] is not None:
        if call["save"]["type"] is None:
            call["save"]["type"] = RunProcessSaveType.STRING

    try:
        call["program"] = BashProgramType(call["program"])
    except ValueError as e:
        raise ValueError(
            f'Program type {call["program"]} is not supported.'
        ) from e


def main() -> None:
    """
    Main function for the test-tool-run-process-plugin.
    """
    print("test-tool-run-process-plugin")



"""
This module contains the JDBC SQL plugin for the universal test tool.
"""
from .main import (
    augment_jdbc_sql_call,
    default_jdbc_sql_call,
    make_jdbc_sql_call,
)

__all__ = [
    "augment_jdbc_sql_call",
    "default_jdbc_sql_call",
    "make_jdbc_sql_call",
]



"""
Module for the assert plugin."""
from enum import Enum
from logging import error, info
from pathlib import Path
from typing import Any, Dict, TypedDict


class Operator(Enum):
    """
    This class represents an operator.
    """

    EQUAL = "=="
    NOT_EQUAL = "!="
    LESS_THAN = "<"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN = ">"
    GREATER_THAN_OR_EQUAL = ">="


class AssertCall(TypedDict):
    """
    This class represents an assert call.
    """

    value: Any
    expected: Any
    operator: Operator
    error_msg: str


# Define the default call
default_assert_call: AssertCall = {
    "value": None,
    "expected": None,
    "operator": "==",  # type: ignore
    "error_msg": "The value is not as expected.",
}

# Define the function that will be called


def make_assert_call(
    call: AssertCall, data: Dict[str, Any]
) -> None:  # pylint: disable=unused-argument
    info(
        f'Asserting {call["value"]} {call["operator"].name} {call["expected"]}'
    )

    # Compare to values dynamically
    if call["operator"] == Operator.EQUAL:
        if call["value"] != call["expected"]:
            error(call["error_msg"])
            assert False, call["error_msg"]
    elif call["operator"] == Operator.NOT_EQUAL:
        if call["value"] == call["expected"]:
            error(call["error_msg"])
            assert False, call["error_msg"]
    elif call["operator"] == Operator.LESS_THAN:
        if call["value"] >= call["expected"]:
            error(call["error_msg"])
            assert False, call["error_msg"]
    elif call["operator"] == Operator.LESS_THAN_OR_EQUAL:
        if call["value"] > call["expected"]:
            error(call["error_msg"])
            assert False, call["error_msg"]
    elif call["operator"] == Operator.GREATER_THAN:
        if call["value"] <= call["expected"]:
            error(call["error_msg"])
            assert False, call["error_msg"]
    elif call["operator"] == Operator.GREATER_THAN_OR_EQUAL:
        if call["value"] < call["expected"]:
            error(call["error_msg"])
            assert False, call["error_msg"]
    else:
        error(f'Unknown operator {call["operator"]}')


def augment_assert_call(
    call: AssertCall, data: Dict, path: Path
) -> None:  # pylint: disable=unused-argument
    """
    This function will be called before the function above.
    It can be used to augment the call.

    Parameters:
    ----------
    call: AssertCall
        The call to augment
    data: Dict
        The data that was passed to the function
    path: Path
        The path of the file that contains the call
    """
    try:
        call["operator"] = Operator(call["operator"])
    except ValueError:
        error(f'Unknown operator {call["operator"]}')


def main() -> None:
    """
    This function will be called when the plugin is loaded.
    """
    print("test-tool-assert-plugin")



"""
CLI interface for test_tool project.

It is executed when the program is called from the command line.
"""
from argparse import ArgumentParser
from logging import DEBUG, INFO, basicConfig
from os import getcwd
from pathlib import Path

from test_tool.base import run_tests


VERSION: str = "0.1.15"

def get_version_from_file() -> str:
    """
    Get the version of the program from the VERSION file.

    Returns:
    --------
    str
        The version of the program.
    """
    file = Path(__file__).parent.joinpath("VERSION")
    with open(file, 'r', encoding="utf-8") as f:
        version = f.read().strip()
    return version


def main() -> None:  # pragma: no cover
    """
    The main function executes on commands:
    `python -m test_tool` and `$ test_tool `.

    This is the program's entry point.
    """
    parser = ArgumentParser(
        prog="test-tool",
        description="This programm runs tests configured by a yaml file.",
        epilog="universal-test-tool Copyright (C) 2023 jackovsky8",
    )

    parser.add_argument(
        "-p",
        "--project",
        action="store",
        help="The path to the project.",
        default=getcwd(),
    )

    parser.add_argument(
        "-ca",
        "--calls",
        action="store",
        help="The filename of the calls configuration.",
        default="calls.yaml",
    )

    parser.add_argument(
        "-d",
        "--data",
        action="store",
        help="The filename of the data configuration.",
        default="data.yaml",
    )

    parser.add_argument(
        "-c",
        "--continue_tests",
        action="store_true",
        help="Continue the tests even if one fails.",
        default=False,
    )

    parser.add_argument(
        "-o",
        "--output",
        action="store",
        help="Create a folder with the logs of the tests.",
        default="runs/%Y%m%d_%H%M%S",
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {VERSION}",
        help="Show the version of the program.",
    )

    parser.add_argument(
        "-X", "--debug", action="store_true", help="Activate debugging."
    )

    args = parser.parse_args()

    log_format: str = "%(asctime)s | %(levelname)s | %(name)s | %(message)s"
    log_level = INFO

    if args.debug:
        log_level = DEBUG
    basicConfig(level=log_level, format=log_format)

    run_tests(
        args.project, args.calls, args.data, args.continue_tests, args.output
    )


if __name__ == "__main__":  # pragma: no cover
    main()



"""
This module contains the REST plugin for the universal test tool.
"""
from .main import augment_rest_call, default_rest_call, make_rest_call

__all__ = ["augment_rest_call", "default_rest_call", "make_rest_call"]



"""
Module for the JDBC SQL plugin.
"""
from logging import debug, error, info
from pathlib import Path
from typing import Any, Dict, List, TypedDict

from jaydebeapi import connect


class JdbcSqlSave(TypedDict):
    row: int
    column: int
    name: str


class JdbcSqlValidate(TypedDict):
    row: int
    column: int
    value: str


class JdbcSqlCall(TypedDict):
    query: str
    save: List[JdbcSqlSave]
    validate: List[JdbcSqlValidate]
    driver: str
    driver_path: str
    url: str
    username: str
    password: str


default_jdbc_sql_call: JdbcSqlCall = {
    "query": None,  # type: ignore
    "save": [],
    "validate": [],
    "driver": "${DB_DRIVER}",
    "driver_path": "${DB_DRIVER_PATH}",
    "url": "${DB_URL}",
    "username": "${DB_USERNAME}",
    "password": "${DB_PASSWORD}",
}


def make_jdbc_sql_call(call: JdbcSqlCall, data: Dict[str, Any]) -> None:
    info(f'Run query: {call["query"]}')

    # Establish the database connection
    debug(f'Connect to {call["url"]} with driver {call["driver"]}')
    with connect(
        call["driver"],
        call["url"],
        [call["username"], call["password"]],
        call["driver_path"],
    ) as conn:
        with conn.cursor() as cursor:
            # Get the query and execute it
            query = call["query"]
            cursor.execute(query)

            # Get the result if needed
            if len(call["save"]) > 0 or len(call["validate"]) > 0:
                rows = cursor.fetchall()

            # Save some values
            for save in call["save"]:
                try:
                    data[save["name"]] = rows[save["row"]][save["column"]]
                except KeyError as e:
                    error(f"Save Object misses key {str(e)}")
                    assert False

            # Validate some values
            for validate in call["validate"]:
                try:
                    if not str(
                        rows[validate["row"]][validate["column"]]
                    ) == str(validate["value"]):
                        error(
                            f'Expected {validate["value"]} in row {validate["row"]}, column {validate["column"]}, but was {rows[validate["row"]][validate["column"]]}'
                        )
                        assert False
                    else:
                        info(
                            f'Validated row {validate["row"]}, column {validate["column"]} with value {rows[validate["row"]][validate["column"]]}'
                        )
                except KeyError as e:
                    error(f"Validate Object misses key {str(e)}")
                    assert False


def augment_jdbc_sql_call(
    call: JdbcSqlCall, data: Dict, path: Path
) -> None:  # pylint: disable=unused-argument
    pass


def main() -> None:
    print("test-tool-jdbc-sql-plugin")



"""
This module contains the SSH command plugin for the universal test tool.
"""
from .main import augment_ssh_cmd_call, default_ssh_cmd_call, make_ssh_cmd_call

__all__ = ["augment_ssh_cmd_call", "default_ssh_cmd_call", "make_ssh_cmd_call"]



"""
test_tool base module.

This is the principal module of the test_tool project.
"""
import sys
from copy import deepcopy
from datetime import datetime
from importlib import import_module
from logging import DEBUG, INFO, FileHandler, Formatter, getLogger
from pathlib import Path
from traceback import print_exception
from types import FunctionType
from typing import Any, Callable, Dict, List, Optional, TypedDict

from yaml import YAMLError, safe_load

test_tool_logger = getLogger("test-tool")


class CallType(TypedDict):
    """
    Call Type.
    """

    default_call: Dict[str, Any]
    augment_call: Callable
    make_call: Callable


# Loaded Plugins
LOADED_CALL_TYPES: Dict[str, CallType] = dict()


class Call(TypedDict):
    """
    Call.
    """

    type: str
    call: Dict[str, Any]
    line: int


# Plugin Name Templates
PLUGIN_NAME_TEMPLATE: str = "test_tool_${plugin}_plugin"

# Plugin Template
PLUGIN_TEMPLATE: Dict[str, str] = {
    "default_call": "default_${plugin}_call",
    "augment_call": "augment_${plugin}_call",
    "make_call": "make_${plugin}_call",
}

PLUGIN_COMPONENT_TYPES: Dict[str, type] = {
    "default_call": dict,
    "augment_call": FunctionType,
    "make_call": FunctionType,
}


def import_plugin(plugin: str) -> bool:
    """
    Dynamically import the specified plugin as a module.

    Parameters
    ----------
    plugin : str
        Name of the plugin.

    Returns
    -------
    bool
        True if the plugin was loaded successfully, False otherwise.
    """
    try:
        plugin_module = import_module(
            PLUGIN_NAME_TEMPLATE.replace("${plugin}", plugin.lower())
        )
    except ModuleNotFoundError:
        test_tool_logger.error(
            "Plugin %s not found, try to install it with pip install"
            + " test_tool_%s_plugin",
            plugin,
            plugin.lower(),
        )
        return False

    # Create a dict for the loaded plugin
    loaded_plugin: CallType = {
        "default_call": {},
        "augment_call": lambda x: x,
        "make_call": lambda x: x,
    }

    # Load the components from the plugin
    to_load: Dict[str, str] = {
        key: val.replace("${plugin}", plugin.lower())
        for key, val in PLUGIN_TEMPLATE.items()
    }
    try:
        for key, component in to_load.items():
            loaded_plugin[key] = getattr(  # type: ignore
                plugin_module, component
            )
            if not isinstance(
                loaded_plugin[key], PLUGIN_COMPONENT_TYPES[key]  # type: ignore
            ):
                test_tool_logger.error(
                    "Module test_tool_%s_plugin is not a valid plugin, %s is"
                    + " not a %s",
                    plugin.lower(),
                    component,
                    PLUGIN_COMPONENT_TYPES[key],
                )
                return False
    except AttributeError:
        test_tool_logger.error(
            "Module test_tool_%s_plugin is not a valid plugin, missing a"
            + " component.",
            plugin.lower(),
        )
        return False

    LOADED_CALL_TYPES[plugin] = loaded_plugin
    return True


def replace_string_variables(
    to_change: str, data: Dict[str, Any]
) -> Optional[str]:
    """
    Replace variables in a string.

    Parameters
    ----------
    to_change : str
        String to change.
    data : Dict[str, Any]
        Data to use for the changes.

    Returns
    -------
    Optional[str]
        The changed string if it was changed, None otherwise.
    """
    changed: str = to_change
    for var, val in data.items():
        # cast to string since we want to replace strings
        val = str(val)

        # The string we want to search for
        origin: str = "${" + var + "}"

        # Check if the string contains the origin
        changed = changed.replace(origin, val)

    if changed != to_change:
        test_tool_logger.debug("Changed value %s to %s.", to_change, changed)
        return changed

    return None


def recursively_replace_variables(
    to_change: Dict[str, Any], data: Dict[str, Any]
) -> Optional[Dict[str, Any]]:
    """
    Recursively replace variables in a dict.

    Parameters
    ----------
    to_change : Dict[str, Any]
        Dict to change.
    data : Dict[str, Any]
        Data to use for the changes.

    Returns
    -------
    Dict[str, Any]
        The changed dict.
    """
    changed: bool = False
    for key, value in to_change.items():
        test_tool_logger.debug("%s: %s", key, value)
        changed = False
        changed_iteration: bool = True
        while changed_iteration:
            changed_iteration = False
            if isinstance(to_change[key], dict):
                changed_iteration = (
                    recursively_replace_variables(value, data) is not None
                )
            elif isinstance(to_change[key], list):
                for idx, val in enumerate(to_change[key]):
                    if isinstance(val, dict):
                        new_val = recursively_replace_variables(val, data)
                        if new_val:
                            changed_iteration = True
                            to_change[key][idx] = new_val
                    elif isinstance(val, str):
                        new_value = replace_string_variables(val, data)
                        if new_value:
                            changed_iteration = True
                            to_change[key][idx] = new_val
            elif isinstance(to_change[key], str):
                new_value = replace_string_variables(to_change[key], data)
                if new_value:
                    changed_iteration = True
                    to_change[key] = new_value

        if changed_iteration:
            changed = True

    if changed:
        return to_change
    else:
        return None


def make_all_calls(
    calls: List[Call], data: Dict[str, Any], path: Path, continue_tests
) -> int:
    """
    Make all calls.

    Parameters
    ----------
    calls : List[Call]
        List of calls.
    data : Dict[str, Any]
        Data to use for the calls.
    path : Path
        Path to the project.
    continue_tests : bool
        Continue tests on error.

    Returns
    -------
    int
        Number of errors.
    """
    errors: int = 0

    # Make the calls and check the response
    for idx, test in enumerate(calls):
        # Check if call type is defined
        try:
            test["type"]
        except KeyError:
            test_tool_logger.debug("No call type defined, using default: REST")
            test["type"] = "REST"

        # Determine if the call type is loaded
        try:
            LOADED_CALL_TYPES[test["type"]]
        except KeyError:
            # Module is not loaded yet, load it
            test_tool_logger.debug(
                "Loading plugin for call type %s", test["type"]
            )
            if not import_plugin(test["type"]):
                test_tool_logger.error(
                    "%s call is not supported", test["type"]
                )
                errors += 1
                continue

        # Merge the default call with the call from the config
        default_call = deepcopy(
            LOADED_CALL_TYPES[test["type"]]["default_call"]
        )
        try:
            call = {**default_call, **test["call"]}
        except KeyError as e:
            if "'call'" == str(e):
                call = default_call
            else:
                test_tool_logger.error("Key %s not found in call", e)
                errors += 1
                continue

        # Recursivly replace variables in call with data
        recursively_replace_variables(call, data)

        # Call the funktion
        try:
            test_tool_logger.info(
                "Make call %s in %s plugin.", idx + 1, test["type"]
            )
            # Augment the call with the data from the config
            LOADED_CALL_TYPES[test["type"]]["augment_call"](call, data, path)
            # Make the call
            LOADED_CALL_TYPES[test["type"]]["make_call"](call, data)
        except AssertionError as e:
            test_tool_logger.error(
                "Assertion error for test from line %s: %s", test["line"], e
            )
            errors += 1
        except Exception as e:  # pylint: disable=broad-except
            test_tool_logger.error(
                'Exception "%s" occured for test from line %s '
                + "(This might be a problem with the plugin or config).",
                e,
                test["line"],
            )
            # if debug:
            if test_tool_logger.getEffectiveLevel() == DEBUG:
                print_exception(type(e), e, e.__traceback__)
            errors += 1

        if errors > 0 and not continue_tests:
            test_tool_logger.error("Stopping on first error")
            break

    return errors


def load_config_yaml(path: Path, add_line_numbers: bool = False) -> Any:
    """
    Load a yaml config file.

    Parameters
    ----------
    path : Path
        Path to the file.
    add_line_numbers : bool, optional
        Add line numbers to the loaded data,
        if the data is a list, by default False

    Returns
    -------
    Dict[str, Any]
        The loaded config.
    """
    with open(path, "r", encoding="utf-8") as file:
        content = file.read()

    try:
        data = safe_load(content)
    except YAMLError as e:
        test_tool_logger.error(e)
        sys.exit(1)

    if add_line_numbers and isinstance(data, list):
        line_numbers = []
        # Every call starts with "- call: or - type:"
        intendation = 0
        for idx, line in enumerate(content.split("\n")):
            if "- call:" in line or "- type:" in line:
                # Count characters before "- call:"
                if idx > 0 and len(line) - len(line.lstrip()) == intendation:
                    line_numbers.append(idx + 1)
                elif idx == 0:
                    intendation = len(line) - len(line.lstrip())
                    line_numbers.append(idx + 1)

        for idx, element in enumerate(data):
            if isinstance(element, dict):
                element["line"] = line_numbers[idx]

    return data


def run_tests(
    project_path_str: str,
    calls_path_str: str,
    data_path_str: str,
    continue_tests: bool,
    output: str,
) -> None:
    """
    Run the tests.

    Parameters
    ----------
    project_path_str : str
        Path to the project.
    calls_path_str : str
        Path to the calls config.
    data_path_str : str
        Path to the data config.
    continue_tests : bool
        Continue tests on error.
    output : str
        Path to the output folder.
    """
    project_path: Path = Path(project_path_str)
    calls_path: Path = project_path.joinpath(calls_path_str)
    data_path: Path = project_path.joinpath(data_path_str)

    test_tool_logger.info(
        "Running tests for project %s", project_path.as_posix()
    )
    test_tool_logger.debug("Calls: %s", calls_path.relative_to(project_path))
    test_tool_logger.debug("Data: %s", data_path.relative_to(project_path))

    # Load the data
    data: Dict[str, Any] = load_config_yaml(data_path)
    if data is None:
        data = {}
    for key, value in data.items():
        test_tool_logger.debug("Loaded %s for %s", value, key)

    # Set the project path in the data
    data["PROJECT_PATH"] = project_path.as_posix()

    # Check if output is set
    if output:
        # Create a string from datetime in format YYYYMMDD_HHMMSS
        now_str = datetime.now().strftime(output)
        output_path: Path = project_path.joinpath(now_str)
        test_tool_logger.debug(
            "Create output folder %s", output_path.relative_to(project_path)
        )
        output_path.mkdir(parents=True, exist_ok=True)

        # Set the output path in the data
        data["OUTPUT_PATH"] = output_path.as_posix()

        # Add Handler to logger to write to file
        fh = FileHandler(output_path.joinpath("run.log"))
        fh.setLevel(INFO)
        fh.setFormatter(
            Formatter("%(asctime)s | %(levelname)s | %(name)s | %(message)s")
        )
        test_tool_logger.addHandler(fh)

    # Load the calls
    calls: List[Call] = load_config_yaml(calls_path, True)
    errors = make_all_calls(calls, data, project_path, continue_tests)

    if errors == 0:
        test_tool_logger.info("Everything OK")
        sys.exit(0)
    else:
        test_tool_logger.error(
            "There occured %s test_tool_logger.errors while testing,"
            + "please check the logs",
            errors,
        )
        sys.exit(1)



"""
Docstring for the main module.
"""
from enum import Enum
from importlib import import_module
from logging import info
from pathlib import Path
from time import sleep
from typing import Any, Callable, Dict, List, Optional, TypedDict, Union

from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.chromium.options import ChromiumOptions
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.wait import WebDriverWait
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import EdgeChromiumDriverManager


def assert_equals(
    reference: Any, given: Any, message: str = "Values should equal:"
) -> None:
    """Assert that the given and reference values are equal."""
    assert (
        reference == given
    ), f"{message} Reference: {reference} Given: {given}"


def assert_not_equals(
    reference: Any, given: Any, message: str = "Values should not equal:"
) -> None:
    """Assert that the given and reference values are not equal."""
    assert (
        reference != given
    ), f"{message} Reference: {reference} Given: {given}"


def assert_true(given: Any, message: str = "Value should be true:") -> None:
    """Assert that the given value is true."""
    assert given, f"{message} Given: {given}"


def assert_false(given: Any, message: str = "Value should be false:") -> None:
    """Assert that the given value is false."""
    assert not given, f"{message} Given: {given}"


def assert_greater_than(
    reference: Any, given: Any, message: str = "Value should be greater than:"
) -> None:
    """Assert that the given value is greater than the reference value."""
    assert (
        reference > given
    ), f"{message} Reference: {reference} Given: {given}"


def assert_greater_than_or_equals(
    reference: Any,
    given: Any,
    message: str = "Value should be greater or equal than:",
) -> None:
    """Assert that the given value is greater than or equals the reference value."""
    assert (
        reference >= given
    ), f"{message} Reference: {reference} Given: {given}"


def assert_less_than(
    reference: Any, given: Any, message: str = "Value should be less than:"
) -> None:
    """Assert that the given value is less than the reference value."""
    assert (
        reference < given
    ), f"{message} Reference: {reference} Given: {given}"


def assert_less_than_or_equals(
    reference: Any,
    given: Any,
    message: str = "Value should be less or equal than:",
) -> None:
    """Assert that the given value is less than or equals the reference value."""
    assert (
        reference <= given
    ), f"{message} Reference: {reference} Given: {given}"


def native_webriver_function(data: str) -> None:
    # import selenium.webdriver programmatically
    print("native_webriver_function")


# Blacklist
BLACKLIST: List[str] = [
    r"__.*__",  # Python magic methods
    "quit",  # Quit the browser
    "close",  # Close the current window
]

ADD_ACTION: Dict[str, Any] = {
    "sleep": sleep,
    "wait": WebDriverWait,
    "native_webriver_function": native_webriver_function,
    "assert_equals": assert_equals,
    "assert_not_equals": assert_not_equals,
    "assert_true": assert_true,
    "assert_false": assert_false,
    "assert_greater_than": assert_greater_than,
    "assert_greater_than_or_equals": assert_greater_than_or_equals,
    "assert_less_than": assert_less_than,
    "assert_less_than_or_equals": assert_less_than_or_equals,
}

# Type definition for arguments
Arguments = List[Union[Dict[Any, Any], List[Any], str, int, float, bool]]


class BrowserType(Enum):
    """
    Enum for the different browser types.
    """

    CHROME = "chrome"
    GECKO = "gecko"
    CHROMIUM_EDGE = "chromium-edge"


# Define the mapping between the browser type and the driver manager
BROWSER_MAPPING: Dict[BrowserType, Callable] = {
    BrowserType.CHROME: ChromeDriverManager,
    BrowserType.GECKO: GeckoDriverManager,
    BrowserType.CHROMIUM_EDGE: EdgeChromiumDriverManager,
}

# Define the mapping between the browser type and the options
BROWSER_OPTIONS_MAPPING: Dict[BrowserType, Callable] = {
    BrowserType.CHROME: ChromeOptions,
    BrowserType.GECKO: FirefoxOptions,
    BrowserType.CHROMIUM_EDGE: ChromiumOptions,
}

# Define the mapping between the browser type and the driver
BROWSER_DRIVER_MAPPING: Dict[BrowserType, Callable] = {
    BrowserType.CHROME: webdriver.Chrome,
    BrowserType.GECKO: webdriver.Firefox,
    BrowserType.CHROMIUM_EDGE: webdriver.ChromiumEdge,
}


class ArgumentType(Enum):
    """
    Enum for the different argument types.
    """

    BOOLEAN = "boolean"
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    LIST = "list"
    DICTIONARY = "dictionary"
    EVAL = "eval"
    ACTION = "action"


class SeleniumCall(TypedDict):
    """
    The structure of a Selenium call.
    """

    actions: List["SeleniumAction"]
    base_url: str
    path: str
    url: Optional[str]
    webdriver: List[BrowserType]


class SeleniumAction(TypedDict):
    """
    The structure of a Selenium action.
    """

    action: str
    args: Optional[Arguments]
    actions: Optional[List["SeleniumAction"]]
    log: Optional[str]


# Define the default call
default_selenium_call: SeleniumCall = {
    "actions": [],
    "base_url": "${GUI_BASE_URL}",
    "path": "${GUI_PATH}",
    "url": None,
    "webdriver": ["chrome"],
}


def import_plugin(plugin: str) -> bool:
    # # Dynamically import the specified components from the module
    # try:
    #     plugin_module = import_module(
    #         PLUGIN_NAME_TEMPLATE.replace("${plugin}", plugin.lower()))
    # except ModuleNotFoundError as e:
    #     error(
    #         f"Plugin {plugin} not found, try to install it with pip install test_tool_{plugin.lower()}_plugin")
    #     return False

    # # Create a dict for the loaded plugin
    # loaded_plugin: Dict[str, Any] = dict()

    # # Load the components from the plugin
    # to_load = [el.replace("${plugin}", plugin.lower())
    #            for el in PLUGIN_TEMPLATE]
    # try:
    #     for component in to_load:
    #         key: str = component.replace(f"{plugin.lower()}_", "")
    #         loaded_plugin[key] = getattr(plugin_module, component)
    #         if type(loaded_plugin[key]) != PLUGIN_COMPONENT_TYPES[key]:
    #             error(
    #                 f"Module test_tool_{plugin.lower()}_plugin is not a valid plugin, {component} is not a {PLUGIN_COMPONENT_TYPES[key]}")
    #             return False
    # except AttributeError as e:
    #     error(
    #         f"Module test_tool_{plugin.lower()}_plugin is not a valid plugin, missing a component.")
    #     return False

    # LOADED_CALL_TYPES[plugin] = loaded_plugin
    return True


def make_selenium_call(call: SeleniumCall, data: Dict[str, Any]) -> None:
    """
    Make a Selenium call.

    Parameters:
    -----------
    call: SeleniumCall
        The call to make.
    data: Dict[str, Any]
        The data to make the call with.
    """
    info(f'Selenium Call wit url {call["url"]}')

    # For each webdriver
    for web_driver in call["webdriver"]:
        # Install the webdriver
        BROWSER_MAPPING[web_driver]().install()

        # Create browser options
        browser_options = BROWSER_OPTIONS_MAPPING[web_driver]()
        for option in call["options"].keys():
            if option in BLACKLIST:
                raise ValueError(f"Option {option} is not allowed.")
            elif option not in dir(browser_options):
                raise ValueError(f"Action {option} is not available.")
            else:
                get_set_or_call_attribute(
                    browser_options,
                    option,
                    get_arguments(call["options"][option], browser_options),
                )

        # Create a new instance of the Chrome driver
        driver = BROWSER_DRIVER_MAPPING[web_driver](options=browser_options)

        # Open the website
        driver.get(call["url"])

        # Run some actions on the website
        for action in call["actions"]:
            run_action(action, driver)

        # Quit the browser
        driver.quit()


def run_action(action: SeleniumAction, element: Any) -> Any:
    """Run an action.

    Parameters:
    - action: The action to run.
    - element: The element to run the action on.

    Returns:
    The result of the action.
    """
    # Log the action
    info(action.get("log", f'Action {action["action"]}'))
    # Action is an extra action that is not part of Selenium
    if action["action"] in ADD_ACTION.keys():
        result = call_callable_with_args(
            ADD_ACTION[action["action"]],
            get_arguments(action.get("args", None), element),
        )
    # Check if the action is in the blacklist
    elif action["action"] in BLACKLIST:
        raise ValueError(f'Action {action["action"]} is not allowed.')
    # Check if the action is available
    elif action["action"] not in dir(element):
        raise ValueError(f'Action {action["action"]} is not available.')
    else:
        # Call the callable with the arguments
        result = get_set_or_call_attribute(
            element,
            action["action"],
            get_arguments(action.get("args", None), element),
        )

    # Get and run the following actions
    actions = action.get("actions", [])
    new_result = result
    for sub_action in actions:
        new_result = run_action(sub_action, result)

    return new_result


def get_arguments(args: Any, element: Any) -> Arguments:
    """Get the arguments of an action.

    Parameters:
    - args: The arguments to get.
    - element: The element to get the arguments for.

    Returns:
    The arguments.
    """
    # List of arguments
    result: Arguments = []
    # Native types
    if (
        isinstance(args, bool)
        or isinstance(args, str)
        or isinstance(args, int)
        or isinstance(args, float)
    ):
        result.append(args)
    # Dict can be everything
    elif isinstance(args, dict):
        arg_type = ArgumentType(args.get("type", None))
        if arg_type in (
            ArgumentType.BOOLEAN,
            ArgumentType.STRING,
            ArgumentType.INTEGER,
            ArgumentType.FLOAT,
            ArgumentType.LIST,
            ArgumentType.DICTIONARY,
        ):
            result.append(args["value"])
        elif arg_type == ArgumentType.EVAL:
            result.append(eval(args["value"]))
        elif arg_type == ArgumentType.ACTION:
            action: SeleniumAction = args["value"]
            result.append(run_action(action, element))
    # List can be a list of everything
    elif isinstance(args, list):
        for arg in args:
            result.extend(get_arguments(arg, element))

    return result


def get_set_or_call_attribute(
    clazz: Any, attribute: str, args: Arguments
) -> Any:
    """Get, set or call an attribute of a class.

    Parameters:
    - clazz: The class to get, set or call the attribute of.
    - attribute: The attribute to get, set or call.
    - args: The arguments to call the attribute with.

    Returns:
    The result of the attribute.
    """
    attribute_of_clazz = getattr(clazz, attribute)
    if callable(attribute_of_clazz):
        return call_callable_with_args(attribute_of_clazz, args)
    else:
        # Get attribute
        if len(args) == 0:
            return getattr(clazz, attribute)
        # Set attribute
        else:
            if isinstance(attribute_of_clazz, bool):
                if not len(args) == 1 or not isinstance(args[0], bool):
                    raise ValueError(
                        f"Option {attribute} is a boolean, but arg isn't."
                    )
                else:
                    setattr(clazz, attribute, args[0])
            elif isinstance(attribute_of_clazz, str):
                if not len(args) == 1 or not isinstance(args[0], str):
                    raise ValueError(
                        f"Option {attribute} is a string, but arg isn't."
                    )
                else:
                    setattr(clazz, attribute, args[0])
            elif isinstance(attribute_of_clazz, int):
                if not len(args) == 1 or not isinstance(args[0], int):
                    raise ValueError(
                        f"Option {attribute} is an integer, but arg isn't."
                    )
                else:
                    setattr(clazz, attribute, args[0])
            elif isinstance(attribute_of_clazz, float):
                if not len(args) == 1 or not isinstance(args[0], float):
                    raise ValueError(
                        f"Option {attribute} is a float, but arg isn't."
                    )
                else:
                    setattr(clazz, attribute, args[0])
            else:
                # Setter no implemented for other values
                raise NotImplementedError(
                    f"Option {attribute} is of type {type(attribute_of_clazz)}."
                )

    return clazz


def call_callable_with_args(
    call: Callable, args: Optional[Union[Dict, List, int, float, str]] = None
) -> Any:
    """Call a callable with the given arguments.

    Parameters:
    - callable: The callable to call.
    - args: The arguments to call the callable with.

    Returns:
    The result of the callable.
    """
    if isinstance(args, dict):
        return call(**args)
    elif isinstance(args, list):
        return call(*args)
    elif args is not None:
        return call(args)
    else:
        return call()


# Define the function that will be called before the function above


def augment_selenium_call(call: SeleniumCall, data: Dict, path: Path) -> None:
    """Augment the call.

    Parameters:
    - call: The call to augment.
    - data: The data to augment the call with.
    - path: The path of the call.

    Returns:
    Nothing.
    """
    # Augment the url
    if call["url"] is None:
        call["url"] = f'{call["base_url"]}{call["path"]}'

    # The webdriver is a list of browser types
    call["webdriver"] = [BrowserType(driver) for driver in call["webdriver"]]


# Define the main function
def main() -> None:
    """
    The main function.^
    """
    print("test-tool-selenium-plugin")



"""
This is the main file of the plugin. It is called by the test tool and
contains the main function.
"""
from enum import Enum
import json
from logging import error, info
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, Tuple, TypedDict
from xml.dom.minidom import parseString

from requests import delete, get, post, put


class Assertion(TypedDict):
    """
    This class represents an assertion.
    """

    value: str | Dict[str, Any] | List[Any]
    only_defined: bool


class Cert(TypedDict):
    """
    This class represents a certificate.
    """

    path: Path
    key: Optional[str]


class BodyType(Enum):
    """
    This class represents the type of the request body.
    """
    TEXT_PLAIN = 'text/plain'
    MULTIPART_FORM_DATA = 'multipart/form-data'
    APPLICATION_JSON = 'application/json'
    APPLICATION_XML = 'application/xml'
    APPLICATION_OCTET_STREAM = 'application/octet-stream'
    APPLICATION_PDF = 'application/pdf'
    APPLICATION_ZIP = 'application/zip'
    APPLICATION_GZIP = 'application/gzip'
    APPLICATION_TAR = 'application/tar'
    APPLICATION_XHTML_XML = 'application/xhtml+xml'
    APPLICATION_XML_DTD = 'application/xml-dtd'
    APPLICATION_XOP_XML = 'application/xop+xml'
    APPLICATION_X_TAR = 'application/x-tar'
    APPLICATION_X_GZIP = 'application/x-gzip'
    APPLICATION_X_BZIP2 = 'application/x-bzip2'
    APPLICATION_X_7Z_COMPRESSED = 'application/x-7z-compressed'
    APPLICATION_X_RAR_COMPRESSED = 'application/x-rar-compressed'
    APPLICATION_X_JAVA_ARCHIVE = 'application/x-java-archive'
    APPLICATION_X_RAR = 'application/x-rar'
    APPLICATION_X_SHOCKWAVE_FLASH = 'application/x-shockwave-flash'
    APPLICATION_X_SQLITE3 = 'application/x-sqlite3'
    APPLICATION_X_WWW_FORM_URLENCODED = 'application/x-www-form-urlencoded'
    AUDIO_MPEG = 'audio/mpeg'
    AUDIO_X_MS_WMA = 'audio/x-ms-wma'
    AUDIO_X_WAV = 'audio/x-wav'
    IMAGE_GIF = 'image/gif'
    IMAGE_JPEG = 'image/jpeg'
    IMAGE_PNG = 'image/png'
    IMAGE_SVG_XML = 'image/svg+xml'
    IMAGE_TIFF = 'image/tiff'
    IMAGE_VND_MICROSOFT_ICON = 'image/vnd.microsoft.icon'
    IMAGE_X_ICON = 'image/x-icon'
    IMAGE_X_MS_BMP = 'image/x-ms-bmp'
    TEXT_CSS = 'text/css'
    TEXT_CSV = 'text/csv'
    TEXT_HTML = 'text/html'
    TEXT_JAVASCRIPT = 'text/javascript'
    TEXT_XML = 'text/xml'
    VIDEO_3GPP = 'video/3gpp'
    VIDEO_3GPP2 = 'video/3gpp2'
    VIDEO_MPEG = 'video/mpeg'
    VIDEO_QUICKTIME = 'video/quicktime'
    VIDEO_X_MSVIDEO = 'video/x-msvideo'
    VIDEO_X_SGI_MOVIE = 'video/x-sgi-movie'
    VIDEO_X_WEBM = 'video/x-webm'
    VIDEO_X_FLV = 'video/x-flv'
    VIDEO_X_MATROSKA = 'video/x-matroska'
    VIDEO_X_MS_WMV = 'video/x-ms-wmv'
    VIDEO_X_MS_ASF = 'video/x-ms-asf'
    VIDEO_X_M4V = 'video/x-m4v'
    VIDEO_X_MSVID = 'video/x-msvid'


class Body(TypedDict):
    """
    This class represents the body of the request.
    """

    type: BodyType
    data: Any


class Method(Enum):
    """
    This class represents the method of the request.
    """

    POST = 1
    GET = 2
    PUT = 3
    DELETE = 4


class Type(Enum):
    """
    This class represents the type of the response.
    """

    XML = 1
    JSON = 2
    TEXT = 3


class File(TypedDict):
    """
    This class represents a file.
    """

    path: str
    binary: bool
    media: str


LoadedFile = Tuple[str, IO[Any], str]
EmptyFile = Tuple[None, str, str]


class RestCall(TypedDict):
    """
    This class represents a rest call.
    """

    # URL
    base_url: str
    path: str
    url: str
    # Request
    method: Method
    body: Optional[Body]
    files: Optional[Dict[str, File | LoadedFile | EmptyFile]]
    multipart: Optional[Dict[str, Tuple[None, str, str]]]
    # payload: Optional[Dict[str, str]]
    # To request
    timeout: int
    headers: Dict
    verify: bool
    cert: Optional[Cert]
    # Verification
    response_type: Type
    assertion: Optional[Assertion]
    hide_logs: bool
    status_codes: List[int]


default_rest_call: RestCall = {
    "base_url": "${REST_BASE_URL}",
    "path": "${REST_PATH}",
    "url": None,  # type: ignore
    "method": "GET",  # type: ignore
    "body": None,
    "files": None,
    "multipart": None,
    # "payload": None,
    # Special
    "timeout": 0,
    "headers": {},
    "verify": False,
    "cert": None,
    # Verification
    "response_type": "JSON",  # type: ignore
    "assertion": None,
    "hide_logs": False,
    "status_codes": [200],
    # url: str | bytes,
    # params: _Params | None = None,
    # *,
    # data: _Data | None = ...,
    # headers: _HeadersMapping | None = ...,
    # cookies: RequestsCookieJar | _TextMapping | None = ...,
    # files: _Files | None = ...,
    # auth: _Auth | None = ...,
    # timeout: _Timeout | None = ...,
    # allow_redirects: bool = ...,
    # proxies: _TextMapping | None = ...,
    # hooks: _HooksInput | None = ...,
    # stream: bool | None = ...,
    # verify: _Verify | None = ...,
    # cert: _Cert | None = ...,
    # json: Any | None = ...
}

default_request_body: Dict[str, Any] = {
    "type": BodyType.TEXT_PLAIN,
    "value": None
}


def pretty_xml(string: str) -> str:
    """
    Return a pretty printed xml string.

    Parameters
    ----------
    xml : str
        The xml string.

    Returns
    -------
    str
        The pretty printed xml string.
    """
    dom = parseString(string)
    return dom.toprettyxml()


def pretty_json(string: str) -> str:
    """
    Return a pretty printed json string.

    Parameters
    ----------
    json : str
        The json string.

    Returns
    -------
    str
        The pretty printed json string.
    """
    return json.dumps(string, indent=2)


# def multipartify(
#     data, parent_key=None, formatter: Optional[Callable] = None
# ) -> dict:
#     if formatter is None:

#         def formatter(v):
#             return (None, v)  # Multipart representation of value

#     if type(data) is not dict:
#         return {parent_key: formatter(data)}

#     converted = []

#     for key, value in data.items():
#         current_key = key if parent_key is None else f"{parent_key}[{key}]"
#         if type(value) is dict:
#             converted.extend(
#                 multipartify(value, current_key, formatter).items()
#             )
#         elif type(value) is list:
#             for ind, list_value in enumerate(value):
#                 iter_key = f"{current_key}[{ind}]"
#                 converted.extend(
#                     multipartify(list_value, iter_key, formatter).items()
#                 )
#         else:
#             converted.append((current_key, formatter(value)))

#     return dict(converted)


def assert_response(call: RestCall) -> None:
    """
    Assert the response of a rest call.

    Parameters
    ----------
    call : RestCall
        The rest call.

    Raises
    ------
    AssertionError
        If the response is not as expected.
    """
    url = call["url"]

    data: Dict = {"headers": call["headers"]}

    # # Add payload as files
    # if call["payload"]:
    #     if call["files"] == None:
    #         call["files"] = {}
    #     for key, value in call["payload"].items():
    #         call["files"][key] = value

    # Add multipart as files
    if call["multipart"]:
        if call["files"] is None:
            call["files"] = {}
        if call["files"] is not None:
            for key, value in call["multipart"].items():
                call["files"][key] = value

    # Add data or files
    if call["files"]:
        data["files"] = call["files"]
        del data["headers"]["Content-Type"]

    # Add timeout
    if call["timeout"]:
        data["timeout"] = call["timeout"]

    # Add verify
    data["verify"] = call["verify"]

    # Add body
    if call["body"] and "data" in call["body"]:
        # Default body type
        if "type" not in call["body"] or not call["body"]["type"]:
            call["body"]["type"] = BodyType.TEXT_PLAIN
        else:
            call["body"]["type"] = BodyType(call["body"]["type"])  # type: ignore

        # Add body
        if call["body"]["type"] == BodyType.APPLICATION_JSON:
            data["json"] = call["body"]["data"]
        elif call["body"]["type"] == BodyType.TEXT_PLAIN:
            data["data"] = call["body"]["data"]
        else:
            raise ValueError("Body type not supported.")

    # url: str | bytes,
    # params: _Params | None = None,
    # *,
    # data: _Data | None = ...,
    # headers: _HeadersMapping | None = ...,
    # cookies: RequestsCookieJar | _TextMapping | None = ...,
    # files: _Files | None = ...,
    # auth: _Auth | None = ...,
    # allow_redirects: bool = ...,
    # proxies: _TextMapping | None = ...,
    # hooks: _HooksInput | None = ...,
    # stream: bool | None = ...,
    # cert: _Cert | None = ...,
    # json: Any | None = ...

    # Make the call
    info(f'Make {call["method"].name} to {url}')
    if call["method"] == Method.GET:
        response = get(url, timeout=10, **data)
    elif call["method"] == Method.POST:
        response = post(url, timeout=10, **data)
    elif call["method"] == Method.PUT:
        response = put(url, timeout=10, **data)
    elif call["method"] == Method.DELETE:
        response = delete(url, timeout=10, **data)

    info(f"Response Status: {response.status_code}")
    if call["hide_logs"] is False:
        try:
            info(f"Response:\n{json.dumps(response.json(), indent=2)}")
        except json.JSONDecodeError:
            info(f'Response: "{response.text}"')

    assert response.status_code in call["status_codes"]

    # Compare the response
    if call["assertion"] is not None:
        if isinstance(call["assertion"]["value"], str):
            assert response.text == call["assertion"]["value"]
        elif isinstance(call["assertion"]["value"], dict):
            if call["assertion"]["only_defined"]:
                response_json = response.json()
                assert isinstance(response_json, dict)
                for key, value in call["assertion"]["value"].items():
                    try:
                        msg = (
                            f'Key "{key}": "{value}" not equal to '
                            + f'"{response_json[key]}".'
                        )
                        assert response_json[key] == value, msg
                    except KeyError:
                        assert False, f'Key "{key}" not found in response.'
            else:
                assert response.json() == call["assertion"]
        elif isinstance(call["assertion"]["value"], list):
            assert response.json() == call["assertion"]
        else:
            error(
                f'Assertion type "{type(call["assertion"])}"'
                + " not supported yet."
            )
            assert False


def make_rest_call(
    call: RestCall, data: Dict[str, Any]  # pylint: disable=unused-argument
) -> None:
    """
    Make a rest call.

    Parameters
    ----------
    call : RestCall
        The rest call.
    data : Dict[str, Any]
        The data from the test tool.

    Raises
    ------
    AssertionError
        If the response is not as expected.
    """
    try:
        assert_response(call)
    except AssertionError as e:
        # Log the expected response
        if call["assertion"] is not None:
            if call["response_type"] == Type.XML:
                error(f'Expexted:\n{pretty_xml(str(call["assertion"]))}')
            elif call["response_type"] == Type.JSON:
                error(f'Expexted:\n{pretty_json(str(call["assertion"]))}')
            else:
                error(f'Expexted:\n{call["assertion"]}')
            raise e


def augment_rest_call(
    call: RestCall, data: Dict, path: Path  # pylint: disable=unused-argument
) -> None:
    """
    Augment a rest call.

    Parameters
    ----------
    call : RestCall
        The rest call.
    data : Dict
        The data from the test tool.
    path : Path
        The project path.
    """
    # Base URL
    if not isinstance(call["base_url"], str):
        raise ValueError("Base URL must be a string.")

    # Path
    if not isinstance(call["path"], str):
        raise ValueError("Path must be a string.")

    # Augment the url
    if call["url"] is None:
        call["url"] = f'{call["base_url"]}{call["path"]}'
    elif not isinstance(call["url"], str):
        raise ValueError("URL must be a string.")

    # Method
    if not isinstance(call["method"], str):
        raise ValueError("Method must be a string.")
    else:
        try:
            call["method"] = Method[call["method"]]
        except KeyError as e:
            raise ValueError("Method is not supported.") from e

    # Data

    # Files
    if call["files"] is not None:
        for key, file in call["files"].items():
            if isinstance(file, dict):
                if file["binary"]:
                    file_type = "rb"
                else:
                    file_type = "r"
                open_file = open(path.joinpath(file["path"]), file_type)
                call["files"][key] = (
                    file["path"],
                    open_file,
                    file["media"],
                )
            else:
                raise ValueError("File is not a dict.")
    # Multipart
    if call["multipart"] is not None:
        for key, part in call["multipart"].items():
            call["multipart"][key] = (
                None, json.dumps(part), "application/json")

    # Payload
    # if call['payload'] is not None:
    #     call["payload"] = multipartify(call["payload"])

    # Timeout
    if not isinstance(call["timeout"], int):
        raise ValueError("Timeout must be an integer, set 0 to disable.")

    # Headers
    if call["headers"] is None:
        raise ValueError("Headers are requiered.")
    elif not isinstance(call["headers"], dict):
        raise ValueError("Headers must be a dict.")

    # Verify
    if not isinstance(call["verify"], bool):
        raise ValueError("Verify must be a boolean.")

    # Cert
    if call["cert"] is not None:
        if call["cert"]["path"] is not None:
            if Path(call["cert"]["path"]).is_absolute():
                call["cert"]["path"] = path.joinpath(call["cert"]["path"])
            else:
                call["cert"]["path"] = path.joinpath(call["cert"]["path"])
        else:
            raise ValueError("Cert path is requiered for cert.")

        if call["cert"]["key"] is not None:
            if not isinstance(call["cert"]["key"], str):
                raise ValueError("Cert key must be a string.")

    # Response type
    if not isinstance(call["response_type"], str):
        raise ValueError("Response type must be a string.")
    else:
        try:
            call["response_type"] = Type[call["response_type"]]
        except KeyError as e:
            raise ValueError("Response type is not supported.") from e

    # Assertion
    if call["assertion"] is not None:
        if not isinstance(call["assertion"], dict):
            raise ValueError("Assertion must be a dict.")

        # Default only_defined
        if ("only_defined" in call["assertion"] and
                call["assertion"]["only_defined"] is not None):
            if not isinstance(call["assertion"]["only_defined"], bool):
                raise ValueError("Only defined must be a boolean.")
        else:
            call["assertion"]["only_defined"] = False

        # Value is requiered
        if ("value" not in call["assertion"] or
                call["assertion"]["value"] is None):
            raise ValueError("Assertion value is requiered.")

        if (not isinstance(call["assertion"]["value"], str)) and (
            not isinstance(call["assertion"]["value"], dict)
        ):
            raise ValueError("Assertion must be a string or a dict.")

    # Hide logs
    if not isinstance(call["hide_logs"], bool):
        raise ValueError("Hide logs must be a boolean.")

    # Status codes
    if not isinstance(call["status_codes"], list):
        raise ValueError("Status codes must be a list.")
    else:
        for status_code in call["status_codes"]:
            if not isinstance(status_code, int):
                raise ValueError("Status codes must be integers.")


def main() -> None:
    """
    The main function of the plugin.
    """
    print("test-tool-rest-plugin")



"""
This module contains the assert plugin for the universal test tool.
"""
from .main import augment_assert_call, default_assert_call, make_assert_call

__all__ = ["augment_assert_call", "default_assert_call", "make_assert_call"]



"""
Module for the sql plus plugin.
"""
from logging import debug, error, info
from pathlib import Path
from subprocess import PIPE, Popen
from threading import Event, Thread
from typing import Any, Dict, List, TypedDict


class SqlPlusNotAvailable(Exception):
    """
    This class represents an exception if the sql plus
    command is not available.
    """


class SqlPlusCall(TypedDict):
    """
    This class represents a sql plus call.
    """

    file: Path
    command: str
    connection: str
    username: str
    password: str


default_sql_plus_call: SqlPlusCall = {
    "file": None,  # type: ignore
    "command": "sqlplus",
    "connection": "${DB_CONNECTION}",
    "username": "${DB_USERNAME}",
    "password": "${DB_PASSWORD}",
}


def log_output(
    pipe,
    prefix: str,
    output_list: List,
    is_error: bool,
    stop_event: Event,
) -> None:
    """
    This function logs the output of the pipe.

    Parameters
    ----------
    pipe : Any
        The pipe to log.
    prefix : str
        The prefix to use.
    output_list : List
        The list to store the output in.
    is_error : bool
        True if the pipe is an error pipe, False otherwise.
    stop_event : Event
        The stop event.
    """
    if is_error:
        log = error
    else:
        log = info

    for line in iter(pipe.readline, b""):
        if line.strip():
            log(f"{prefix}: {line.strip()}")
            output_list.append(line)
        # Stop the thread if the stop event is set
        if stop_event and stop_event.is_set():
            break


def run_cmd(command: str) -> str:
    """
    This function runs the command.

    Parameters
    ----------
    command : str
        The command to run.

    Returns
    -------
    str
        The output of the command.

    Raises
    ------
    AssertionError
        If the return code is not 0.
    """
    process = Popen(
        command, shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE, text=True
    )
    debug(f"Run command: {command}")

    # Create an event to signal threads to stop
    stop_event = Event()

    # Start threads to read output and error streams
    output_list: List[str] = []
    output_thread = Thread(
        target=log_output,
        args=(process.stdout, "Output", output_list, False, stop_event),
    )
    error_thread = Thread(
        target=log_output,
        args=(process.stderr, "Error", output_list, True, stop_event),
    )

    output_thread.start()
    error_thread.start()

    return_code = process.wait()
    assert (
        return_code == 0
    ), f"Sqlplus command failed with return code {return_code}"

    # Signal the threads to stop
    stop_event.set()

    # Wait for the threads to finish
    output_thread.join()
    error_thread.join()

    return "".join(output_list).strip()


def check_sql_plus(command: str) -> None:
    """
    This function checks if the sql plus command is available.

    Parameters
    ----------
    command : str
        The sql plus command to check.

    Raises
    ------
    SqlPlusNotAvailable
        If the sql plus command is not available.
    """
    command = f"{command} -v"
    try:
        result = run_cmd(command)
        if result.startswith("SQL*Plus:"):
            return
    except AssertionError:
        pass

    raise SqlPlusNotAvailable("Sql plus command not available.")


def run_sql(
    command: str, file: Path, connection: str, username: str, password: str
) -> None:
    """
    This function runs the sql plus command.
    """
    command = f'echo quit | {command} -S {username}/{password}@"{connection}" @{file}'
    run_cmd(command)


def make_sql_plus_call(call: SqlPlusCall, data: Dict[str, Any]) -> None:
    """
    This function makes the sql plus call.

    Parameters
    ----------
    call : SqlPlusCall
        The sql plus call to make.
    data : Dict[str, Any]
        The data dictionary.
    path : Path
        The path to the test file.
    """
    info(f'Run sql file: {call["file"]}')

    # Check if the sql plus command is available
    check_sql_plus(call["command"])

    # Run the sql plus command
    run_sql(
        call["command"],
        call["file"],
        call["connection"],
        call["username"],
        call["password"],
    )


def augment_sql_plus_call(
    call: SqlPlusCall, data: Dict[str, Any], path: Path
) -> None:
    """
    This function augments the sql plus call with the default values.

    Parameters
    ----------
    call : SqlPlusCall
        The sql plus call to augment.
    data : Dict[str, Any]
        The data dictionary.
    path : Path
        The path to the test file.
    """
    if call["file"] is None:
        raise ValueError("The file parameter is required.")
    else:
        call["file"] = path.joinpath(call["file"])


def main() -> None:
    """
    This is the main entry point of the plugin.
    """
    print("test-tool-sql-plus-plugin")



"""
This module contains the plugin for reading the manifest of a jar file.
"""
from .main import (
    augment_read_jar_manifest_call,
    default_read_jar_manifest_call,
    make_read_jar_manifest_call,
)

__all__ = [
    "augment_read_jar_manifest_call",
    "default_read_jar_manifest_call",
    "make_read_jar_manifest_call",
]



"""
This module contains the run process plugin for the universal test tool.
"""
from .main import (
    augment_run_process_call,
    default_run_process_call,
    make_run_process_call,
)

__all__ = [
    "augment_run_process_call",
    "default_run_process_call",
    "make_run_process_call",
]



"""
This module contains the copy files over SSH plugin for the 
universal test tool.
"""
from .main import (
    augment_copy_files_ssh_call,
    default_copy_files_ssh_call,
    make_copy_files_ssh_call,
)

__all__ = [
    "augment_copy_files_ssh_call",
    "default_copy_files_ssh_call",
    "make_copy_files_ssh_call",
]



"""
This module contains the Selenium plugin for the universal test tool.
"""
from .main import (
    augment_selenium_call,
    default_selenium_call,
    make_selenium_call,
)

__all__ = [
    "augment_selenium_call",
    "default_selenium_call",
    "make_selenium_call",
]



"""
Module for copying files over SSH.
"""
from logging import debug, error, info
from pathlib import Path
from stat import S_ISDIR
from typing import Any, Callable, Dict, TypedDict

from paramiko import AutoAddPolicy, SSHClient


class CopyFilesSshCall(TypedDict):
    user: str
    password: str
    host: str
    local_path: Path
    remote_path: Path
    download: bool


default_copy_files_ssh_call: CopyFilesSshCall = {
    "user": "${REMOTE_CMD_USER}",
    "password": "${REMOTE_CMD_PASSWORD}",
    "host": "${REMOTE_CMD_HOST}",
    "local_path": None,  # type: ignore
    "remote_path": None,  # type: ignore
    "download": False,
}


def run_with_ssh_client(
    user: str, host: str, password: str, call: Callable[[SSHClient], None]
) -> None:
    # Create an SSH client
    info(f"Connect to {user}@{host}")
    client = SSHClient()

    # Automatically add the server's host key
    client.set_missing_host_key_policy(AutoAddPolicy())

    try:
        # Connect to the remote server
        client.connect(host, username=user, password=password)

        # Run the callable
        call(client)
    finally:
        # Close the SSH connection
        client.close()


def upload_directory(local_path: Path, remote_path: Path, sftp):
    debug(
        f"Upload directory {local_path.as_posix()} to {remote_path.as_posix()}"
    )
    try:
        # Create the remote directory if it doesn't exist
        try:
            sftp.stat(remote_path.as_posix())
        except FileNotFoundError:
            sftp.mkdir(remote_path.as_posix())

        # Upload the files in the local directory
        for item in local_path.iterdir():
            if item.is_file():
                sftp.put(
                    item.as_posix(), remote_path.joinpath(item.name).as_posix()
                )
            elif item.is_dir():
                upload_directory(remote_path.joinpath(item.name), item, sftp)
    except FileNotFoundError as e:
        raise ValueError(f"Local directory '{local_path}' not found.") from e
    except Exception as e:
        print(f"Error uploading directory: {e}")


def download_directory(remote_path: Path, local_path: Path, sftp):
    debug(
        f"Download directory {remote_path.as_posix()} to {local_path.as_posix()}"
    )
    for item in sftp.listdir_attr(remote_path.as_posix()):
        remote_item = remote_path.joinpath(item.filename)
        local_item = local_path.joinpath(item.filename)

        if S_ISDIR(item.st_mode):
            local_item.mkdir(parents=True, exist_ok=True)
            download_directory(remote_item, local_item, sftp)
        else:
            sftp.get(remote_item.as_posix(), local_item.as_posix())


def copy_remote_file(
    client: SSHClient,
    local_path: Path,
    remote_path: Path,
    download: bool = False,
) -> None:
    # Use SFTP to copy the file
    with client.open_sftp() as sftp:
        if not download:
            info(
                f"Copy the file {local_path.as_posix()} to {remote_path.as_posix()}"
            )
            # Create the remote directory if it doesn't exist
            try:
                sftp.stat(remote_path.as_posix())
            except FileNotFoundError:
                sftp.mkdir(remote_path.as_posix())

            # Copy the file
            if local_path.is_dir():
                error("Can not copy a directory from local to remote")
                # TODO: Test
                # upload_directory(remote_path, local_path, sftp)
                assert False
            else:
                sftp.put(
                    local_path.as_posix(),
                    remote_path.joinpath(local_path.name).as_posix(),
                )
        else:
            info(
                f"Copy the file {remote_path.as_posix()} to {local_path.as_posix()}"
            )
            # Create the local directory if it doesn't exist
            local_path.mkdir(parents=True, exist_ok=True)

            # Copy the file
            if S_ISDIR(sftp.stat(remote_path.as_posix()).st_mode):
                download_directory(remote_path, local_path, sftp)
            else:
                sftp.get(
                    remote_path.as_posix(),
                    local_path.joinpath(remote_path.name).as_posix(),
                )


def make_copy_files_ssh_call(
    call: CopyFilesSshCall,
    data: Dict[str, Any],  # pylint: disable=unused-argument
) -> None:
    # Run the cmd with client
    run_with_ssh_client(
        call["user"],
        call["host"],
        call["password"],
        lambda client: copy_remote_file(
            client, call["local_path"], call["remote_path"], call["download"]
        ),
    )


def augment_copy_files_ssh_call(
    call: CopyFilesSshCall,
    data: Dict,
    path: Path,  # pylint: disable=unused-argument
) -> None:
    if call["local_path"] is not None:
        call["local_path"] = Path(call["local_path"])

    if not call["local_path"].is_absolute():
        call["local_path"] = path.joinpath(call["local_path"])

    if call["remote_path"] is not None:
        call["remote_path"] = Path(call["remote_path"])


def main() -> None:
    print("test-tool-copy-files-ssh-plugin")



"""
The module contains the plugin for the universal test tool.
"""
from .main import (
    augment_sql_plus_call,
    default_sql_plus_call,
    make_sql_plus_call,
)

__all__ = [
    "augment_sql_plus_call",
    "default_sql_plus_call",
    "make_sql_plus_call",
]



"""
This is a plugin for the universal test tool.

It provides the ability to run commands on a remote host via SSH.
"""
import time
import datetime
from enum import Enum
from logging import info
from pathlib import Path
from typing import Any, Dict, TypedDict


class Action(Enum):
    """
    Class for Action.
    """

    START = "start"
    STOP = "stop"


class TimingCall(TypedDict):
    """
    Class for TimingCall.
    """

    name: str
    action: Action


class TimingPluginData(TypedDict):
    """
    Class for TimingPluginData.
    """

    start_time: float
    stop_time: float
    duration: float


default_timing_call: TimingCall = {
    "name": "TIMER",
    "action": "start",  # type: ignore
}


def make_timing_call(call: TimingCall, data: Dict[str, Any]) -> None:
    """
    Make an SSH command call.

    Parameters
    ----------
    call : TimingCall
        The call.
    data : Dict[str, Any]
        The data.
    """
    t_data: TimingPluginData = data["timing_plugin_calls"][call["name"]]
    if call["action"] == Action.START:
        info(f'Start timer {call["name"]}')
        t_data["start_time"] = int(time.time() * 1e9)
    elif call["action"] == Action.STOP:
        info(f'Stop timer {call["name"]}')
        t_data["stop_time"] = int(time.time() * 1e9)
        t_data["duration"] = t_data["stop_time"] - t_data["start_time"]

        data[f'TIMING_{call["name"]}_START_NANO'] = t_data["start_time"]
        data[f'TIMING_{call["name"]}_STOP_NANO'] = t_data["stop_time"]
        data[f'TIMING_{call["name"]}_DURATION_NANO'] = t_data["duration"]

        data[f'TIMING_{call["name"]}_START'] = datetime.datetime.utcfromtimestamp(
            t_data["start_time"] / 1e9).strftime("%Y-%m-%d %H:%M:%S.%f UTC")
        data[f'TIMING_{call["name"]}_STOP'] = datetime.datetime.utcfromtimestamp(
            t_data["stop_time"] / 1e9).strftime("%Y-%m-%d %H:%M:%S.%f UTC")
        data[f'TIMING_{call["name"]}_DURATION'] = datetime.datetime.utcfromtimestamp(
            t_data["duration"] / 1e9).strftime("%H:%M:%S.%f")


def augment_timing_call(call: TimingCall, data: Dict, path: Path) -> None:  # pylint: disable=unused-argument
    """
    Augment an Timing command call.

    Parameters
    ----------
    call : TimingCall
        The call.
    data : Dict
        The data.
    path : Path
        The path.
    """
    if "name" not in call:
        raise ValueError("The name is missing in the call.")
    else:
        call["name"] = str(call["name"]).upper()

    if "action" not in call:
        raise ValueError("The action is missing in the call.")
    else:
        call["action"] = Action(call["action"])  # type: ignore

    if "timing_plugin_calls" not in data:
        data["timing_plugin_calls"] = {}
    
    data["timing_plugin_calls"][call["name"]] = {
        "start_time": 0.0,
        "stop_time": 0.0,
        "duration": 0.0,
    }


def main() -> None:
    """
    Main function.
    """
    print("test-tool-timing-plugin")



"""Python setup.py for test_tool package"""
import io
import os
from setuptools import find_packages, setup


def read(*paths, **kwargs):
    """Read the contents of a text file safely.
    >>> read("test_tool", "VERSION")
    '0.1.0'
    >>> read("README.md")
    ...
    """

    content = ""
    with io.open(
        os.path.join(os.path.dirname(__file__), *paths),
        encoding=kwargs.get("encoding", "utf8"),
    ) as open_file:
        content = open_file.read().strip()
    return content


def read_requirements(path):
    return [
        line.strip()
        for line in read(path).split("\n")
        if not line.startswith(('"', "#", "-", "git+"))
    ]


setup(
    name="universal-test-tool",
    version=read("test_tool", "VERSION"),
    description="Awesome universal-test-tool to make tests configurable with a yaml file.",
    url="https://github.com/jackovsky8/universal-test-tool/",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="jackovsky8",
    packages=find_packages(exclude=["tests", ".github"]),
    install_requires=read_requirements("requirements.txt"),
    entry_points={
        "console_scripts": [
            "test-tool = test_tool.__main__:main",
            "test-tool-assert-plugin = test_tool_assert_plugin.__main__:main",
            "test-tool-copy-files-ssh-plugin = test_tool_copy_files_ssh_plugin.__main__:main",
            "test-tool-jdbc-sql-plugin = test_tool_jdbc_sql_plugin.__main__:main",
            "test-tool-read-jar-manifest-plugin = test_tool_read_jar_manifest_plugin.__main__:main",
            "test-tool-rest-plugin = test_tool_rest_plugin.__main__:main",
            "test-tool-run-process-plugin = test_tool_run_process_plugin.__main__:main",
            "test-tool-selenium-plugin = test_tool_selenium_plugin.__main__:main",
            "test-tool-sql-plus-plugin = test_tool_sql_plus_plugin.__main__:main",
            "test-tool-ssh-cmd-plugin = test_tool_ssh_cmd_plugin.__main__:main",
            "test-tool-timing-plugin = test_tool_timing_plugin.__main__:main",
        ]
    },
    extras_require={"test": read_requirements("requirements-test.txt")},
)



"""
Module for reading a JAR manifest file.
"""
from logging import debug, error, info
from pathlib import Path
from typing import Any, Dict, List, Optional, TypedDict
from zipfile import ZipFile


class SaveValue(TypedDict):
    """
    This class represents a save value.
    """

    name: str
    key: Optional[str]


class ReadJarManifestCall(TypedDict):
    """
    This class represents a read JAR manifest call.
    """

    jar_path: Path
    manifest_path: str
    save: List[SaveValue]


default_read_jar_manifest_call: ReadJarManifestCall = {
    "jar_path": "${PATH_JAR_FILE}",  # type: ignore
    "manifest_path": "META-INF/MANIFEST.MF",
    "save": [],
}


def make_read_jar_manifest_call(
    call: ReadJarManifestCall, data: Dict[str, Any]
) -> None:
    """
    This function makes the call to read the JAR manifest.

    Parameters
    ----------
    call : ReadJarManifestCall
        The call to make.
    data : Dict[str, Any]
        The data from the test tool.
    """
    info(f'Reading manifest from {call["jar_path"]}:{call["manifest_path"]}')

    with ZipFile(call["jar_path"], "r") as zip_file:
        with zip_file.open(call["manifest_path"]) as f:
            manifest_content: str = f.read().decode("utf-8")

    # Parse specific attributes from the manifest
    # Every line is a key-value pair
    manifest_lines: List[str] = manifest_content.split("\r\n")
    manifest: Dict[str, str] = {}
    for line in manifest_lines:
        if ": " in line:
            key, value = line.split(": ", 1)
            if isinstance(key, str) and isinstance(value, str):
                manifest[key] = value

    info(f"Got manifest:\n{manifest}")

    # Save specific attributes from the manifest
    for save in call["save"]:
        if save["key"] in manifest:
            data[save["name"]] = manifest[save["key"]]

            debug(f'Saving {save["name"]}={data[save["name"]]}')
        else:
            error(f'No value for {save["name"]}')


def augment_read_jar_manifest_call(
    call: ReadJarManifestCall,
    data: Dict,
    path: Path,  # pylint: disable=unused-argument
) -> None:
    """
    This function augments the call with the absolute path to the JAR file.

    Parameters
    ----------
    call : ReadJarManifestCall
        The call to augment.
    data : Dict
        The data from the test tool.
    path : Path
        The path of the test tool.
    """
    if call["jar_path"] is None:
        raise ValueError("jar_path is required")
    else:
        call["jar_path"] = Path(call["jar_path"])
        if not call["jar_path"].is_absolute():
            call["jar_path"] = path.joinpath(call["jar_path"])


def main() -> None:
    """
    This is the main function of the plugin.
    """
    print("test-tool-read-jar-manifest-plugin")



"""
This is a plugin for the universal test tool.

It provides the ability to run commands on a remote host via SSH.
"""
from logging import error, info
from pathlib import Path
from typing import Any, Callable, Dict, TypedDict

from paramiko import AutoAddPolicy, SSHClient


class SshCmdCall(TypedDict):
    """
    Class for SshCmdCall.
    """

    user: str
    password: str
    host: str
    cmd: str
    return_code: int


default_ssh_cmd_call: SshCmdCall = {
    "user": "${REMOTE_CMD_USER}",
    "password": "${REMOTE_CMD_PASSWORD}",
    "host": "${REMOTE_CMD_HOST}",
    "cmd": "",
    "return_code": 0,
}


def run_with_ssh_client(
    user: str, host: str, password: str, call: Callable[[SSHClient], None]
) -> None:
    """
    Run the callable with an SSH client.

    Parameters
    ----------
    user : str
        The user name.
    host : str
        The host name.
    password : str
        The password.
    call : Callable[[SSHClient], None]
        The callable.
    """
    # Create an SSH client
    info(f"Connect to {user}@{host}")
    client = SSHClient()

    # Automatically add the server's host key
    client.set_missing_host_key_policy(AutoAddPolicy())

    try:
        # Connect to the remote server
        client.connect(host, username=user, password=password)

        # Run the callable
        call(client)
    finally:
        # Close the SSH connection
        client.close()


def run_ssh_cmd(
    client: SSHClient, cmd: str, expected_return_code: int
) -> None:
    """
    Run an SSH command.

    Parameters
    ----------
    client : SSHClient
        The SSH client.
    cmd : str
        The command.
    expected_return_code : int
        The expected return code.
    """
    # Execute the command
    _, stdout, stderr = client.exec_command(cmd)

    # Read and print the output
    output_str = stdout.read().decode("utf-8")
    error_str = stderr.read().decode("utf-8")
    if output_str:
        info(output_str.strip().strip("'").strip('"'))
    if error_str:
        error(error_str.strip().strip("'").strip('"'))

    # Check the return code
    return_code = stdout.channel.recv_exit_status()
    if expected_return_code is not None:
        equals: bool = return_code == expected_return_code
        assert (
            equals
        ), f"Expected return code {expected_return_code}, got {return_code}"
    else:
        info(f"SSH Command return with code: {return_code}")


def make_ssh_cmd_call(call: SshCmdCall, data: Dict[str, Any]) -> None:
    """
    Make an SSH command call.

    Parameters
    ----------
    call : SshCmdCall
        The call.
    data : Dict[str, Any]
        The data.
    """
    info(f'Run the cmd {call["cmd"]} remotely.')
    # Run the cmd with client
    run_with_ssh_client(
        call["user"],
        call["host"],
        call["password"],
        lambda client: run_ssh_cmd(client, call["cmd"], call["return_code"]),
    )


def augment_ssh_cmd_call(call: SshCmdCall, data: Dict, path: Path) -> None:
    """
    Augment an SSH command call.

    Parameters
    ----------
    call : SshCmdCall
        The call.
    data : Dict
        The data.
    path : Path
        The path.
    """
    if call["return_code"] is not None:
        call["return_code"] = int(call["return_code"])


def main() -> None:
    """
    Main function.
    """
    print("test-tool-ssh-cmd-plugin")



"""
This module contains the Timing command plugin for the universal test tool.
"""
from .main import augment_timing_call, default_timing_call, make_timing_call

__all__ = ["augment_timing_call", "default_timing_call", "make_timing_call"]
